from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from functools import partial
from io import StringIO
from io import TextIOWrapper
from multiprocessing import cpu_count
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast

from pqdm import processes

from dycw_utilities.tqdm import tqdm


_T = TypeVar("_T")


def pmap(
    func: Callable[..., _T],
    /,
    *iterables: Iterable[Any],
    parallelism: Literal["processes", "threads"] = "processes",
    n_jobs: Optional[int] = None,
    bounded: bool = False,
    exception_behaviour: Literal["ignore", "immediate", "deferred"] = "ignore",
    desc: Optional[str] = None,
    total: Optional[Union[int, float]] = None,
    leave: Optional[bool] = True,
    file: Optional[Union[TextIOWrapper, StringIO]] = None,
    ncols: Optional[int] = None,
    mininterval: Optional[float] = 0.1,
    maxinterval: Optional[float] = 10,
    miniters: Optional[Union[int, float]] = None,
    ascii: Optional[Union[bool, str]] = None,
    unit: Optional[str] = "it",
    unit_scale: Optional[Union[bool, int, str]] = False,
    dynamic_ncols: Optional[bool] = False,
    smoothing: Optional[float] = 0.3,
    bar_format: Optional[str] = None,
    initial: Optional[Union[int, float]] = 0,
    position: Optional[int] = None,
    postfix: Optional[Mapping[str, Any]] = None,
    unit_divisor: Optional[float] = 1000,
    write_bytes: Optional[bool] = None,
    lock_args: Optional[tuple[Any, ...]] = None,
    nrows: Optional[int] = None,
    colour: Optional[str] = None,
    delay: Optional[float] = 0,
    gui: Optional[bool] = False,
    **kwargs: Any,
) -> list[_T]:
    """Parallel map, powered by `pqdm`."""

    return pstarmap(
        func,
        zip(*iterables),
        parallelism=parallelism,
        n_jobs=n_jobs,
        bounded=bounded,
        exception_behaviour=exception_behaviour,
        desc=desc,
        total=total,
        leave=leave,
        file=file,
        ncols=ncols,
        mininterval=mininterval,
        maxinterval=maxinterval,
        miniters=miniters,
        ascii=ascii,
        unit=unit,
        unit_scale=unit_scale,
        dynamic_ncols=dynamic_ncols,
        smoothing=smoothing,
        bar_format=bar_format,
        initial=initial,
        position=position,
        postfix=postfix,
        unit_divisor=unit_divisor,
        write_bytes=write_bytes,
        lock_args=lock_args,
        nrows=nrows,
        colour=colour,
        delay=delay,
        gui=gui,
        **kwargs,
    )


def pstarmap(
    func: Callable[..., _T],
    iterable: Iterable[tuple[Any, ...]],
    /,
    *,
    parallelism: Literal["processes", "threads"] = "processes",
    n_jobs: Optional[int] = None,
    bounded: bool = False,
    exception_behaviour: Literal["ignore", "immediate", "deferred"] = "ignore",
    desc: Optional[str] = None,
    total: Optional[Union[int, float]] = None,
    leave: Optional[bool] = True,
    file: Optional[Union[TextIOWrapper, StringIO]] = None,
    ncols: Optional[int] = None,
    mininterval: Optional[float] = 0.1,
    maxinterval: Optional[float] = 10,
    miniters: Optional[Union[int, float]] = None,
    ascii: Optional[Union[bool, str]] = None,
    unit: Optional[str] = "it",
    unit_scale: Optional[Union[bool, int, str]] = False,
    dynamic_ncols: Optional[bool] = False,
    smoothing: Optional[float] = 0.3,
    bar_format: Optional[str] = None,
    initial: Optional[Union[int, float]] = 0,
    position: Optional[int] = None,
    postfix: Optional[Mapping[str, Any]] = None,
    unit_divisor: Optional[float] = 1000,
    write_bytes: Optional[bool] = None,
    lock_args: Optional[tuple[Any, ...]] = None,
    nrows: Optional[int] = None,
    colour: Optional[str] = None,
    delay: Optional[float] = 0,
    gui: Optional[bool] = False,
    **kwargs: Any,
) -> list[_T]:
    """Parallel map, powered by `pqdm`."""

    n_jobs = _get_n_jobs(n_jobs)
    tqdm_class = cast(Any, tqdm)
    if parallelism == "processes":
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **({} if desc is None else {"desc": desc}),
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    else:
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **({} if desc is None else {"desc": desc}),
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    return list(result)


def _get_n_jobs(n_jobs: Optional[int], /) -> int:
    if (n_jobs is None) or (n_jobs <= 0):
        return cpu_count()  # pragma: no cover
    else:
        return n_jobs


def _starmap_helper(func: Callable[..., _T], *args: Any) -> _T:
    return func(*args)



import re
from collections.abc import Callable
from typing import Any

from fastapi import APIRouter as _APIRouter
from fastapi.types import DecoratedCallable


_PATTERN = re.compile(r"(^/$)|(^.+[^\/]$)")


class APIRouter(_APIRouter):
    def api_route(  # type: ignore
        self, *, path: str, include_in_schema: bool = True, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        if _PATTERN.search(path):
            return super().api_route(
                path, include_in_schema=include_in_schema, **kwargs
            )
        else:
            raise ValueError(f"Invalid route: {path}")



from collections.abc import Iterable
from collections.abc import Mapping
from io import StringIO
from io import TextIOWrapper
from typing import Any
from typing import Optional
from typing import Union

from tqdm import tqdm as _tqdm

from dycw_utilities.pytest import is_pytest


class tqdm(_tqdm):
    def __init__(
        self,
        iterable: Optional[Iterable[Any]] = None,
        desc: Optional[str] = None,
        total: Optional[Union[int, float]] = None,
        leave: Optional[bool] = True,
        file: Optional[Union[TextIOWrapper, StringIO]] = None,
        ncols: Optional[int] = None,
        mininterval: Optional[float] = 0.1,
        maxinterval: Optional[float] = 10,
        miniters: Optional[Union[int, float]] = None,
        ascii: Optional[Union[bool, str]] = None,
        unit: Optional[str] = "it",
        unit_scale: Optional[Union[bool, int, str]] = False,
        dynamic_ncols: Optional[bool] = False,
        smoothing: Optional[float] = 0.3,
        bar_format: Optional[str] = None,
        initial: Optional[Union[int, float]] = 0,
        position: Optional[int] = None,
        postfix: Optional[Mapping[str, Any]] = None,
        unit_divisor: Optional[float] = 1000,
        write_bytes: Optional[bool] = None,
        lock_args: Optional[tuple[Any, ...]] = None,
        nrows: Optional[int] = None,
        colour: Optional[str] = None,
        delay: Optional[float] = 0,
        gui: Optional[bool] = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            iterable=iterable,
            desc=desc,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=is_pytest(),
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )



from typing import Any
from typing import Optional

from hypothesis.strategies import SearchStrategy
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine

from dycw_utilities.hypothesis import draw_and_map
from dycw_utilities.hypothesis.tempfile import temp_dirs
from dycw_utilities.sqlalchemy import create_engine
from dycw_utilities.tempfile import TemporaryDirectory


def sqlite_engines(
    *, metadata: Optional[MetaData] = None, base: Any = None
) -> SearchStrategy[Engine]:
    """Strategy for generating SQLite engines."""

    return draw_and_map(
        _draw_sqlite_engines, temp_dirs(), metadata=metadata, base=base
    )


def _draw_sqlite_engines(
    temp_dir: TemporaryDirectory,
    /,
    *,
    metadata: Optional[MetaData] = None,
    base: Any = None,
) -> Engine:
    path = temp_dir.name.joinpath("db.sqlite")
    engine = create_engine("sqlite", database=path.as_posix())
    if metadata is not None:
        metadata.create_all(engine)
    if base is not None:
        base.metadata.create_all(engine)

    # attach temp_dir to the engine, so as to keep it alive
    engine.temp_dir = temp_dir  # type: ignore

    return engine
